exception Kill_connection;;

let default_port = "80";;

let error err mes =
  prerr_endline (String.concat " " [ "Error:"; err; mes]);
  Pervasives.exit 2;;

let error_info err mes =
  prerr_endline (String.concat " " [ "Error (Non-Fatal to Server):"; err; mes]);;
    

(*******************************************************************)

(* Parsing expressions and functions *)

type regexp = { regexp : Str.regexp; fields : (int * string option) list; }

let unsafe_regexp_match r string =
  let get (pos, default) =
    try Str.matched_group pos string
    with Not_found ->
      match default with Some s -> s
      | _ -> raise Not_found in
  try
    if Str.string_match r.regexp string 0 then
      Some (List.map get r.fields)
    else None
  with Not_found -> None;;

let strlock = Mutex.create();;

let regexp_match r string =
  Mutex.lock strlock;
  try let v = unsafe_regexp_match r string in Mutex.unlock strlock; v
    with x -> Mutex.unlock strlock; raise x;;

let host_regexp =
  { regexp = Str.regexp "\\([^/:]*\\)\\(:\\([0-9]+\\)\\)?";
    fields = [ 1, None; 3, Some default_port; ] };;

let chunk_size_regexp =
  { regexp = Str.regexp "\\([^ \t]*\\)\\([ \t]*\\)\r";
    fields = [ 1, None  ] };;

let host_regexp_2 =
  { regexp = Str.regexp "\\([^/:]*\\)\\(:\\([0-9]+\\)\\)?";
    fields = [ 1, None; 2, Some "" ] };;

let url_regexp =
  { regexp = Str.regexp "http://\\([^/:]*\\(:[0-9]+\\)?\\)\\(/.*\\)";
    fields = [ 1, None; 3, None ] };;

let get_regexp =
  { regexp = Str.regexp "^[Gg][Ee][Tt][ \t]+\\(.*[^ \t]\\)[ \t]*\r";
    fields = [ 1, None ] }

let request_regexp_only_reqline=
  { regexp = Str.regexp "\\(.*\\)\r\\(\\(.*\r\\)*\\)\r";
    fields = [ 1, None; 2, None ] }

let request_regexp =
  { regexp = Str.regexp "\\(.*\\)\r\n\\(\\(\\(.*\\)\r\n\\)*\\)\r\n\\(\\(.*\n\\)*.*\\)";
    fields = [ 1, None; 2, Some ""; 5, Some "Empty" ] }

let request_line_regexp =
  { regexp = Str.regexp "\\([^ \t]*\\)[ \t]+\\([^ \t]*\\)[ \t]*\\([^ \t]+\\)?[ \t]*\r";
    fields = [ 1, None; 2, None; 3, Some "HTTP/0.9" ] }

let status_line_regexp =
  { regexp = Str.regexp "\\([^ \t]*\\)[ \t]+\\([0-9][0-9][0-9]\\)[ \t]+\\(.*\\)\r";
    fields = [ 1, None; 2, None; 3, Some "No Reason - non fatal error" ] }

let host_header_regexp =
  { regexp = Str.regexp "\\(\\(.*\r\n\\)*\\)\\([Hh][Oo][sS][tT][:].*\r\n\\)\\(\\(.*\r\n\\)*\\)";
    fields = [ 1, None; 4, None ] }

let content_length_regexp =
  { regexp = Str.regexp "\\(\\(.*\r\n\\)*\\)\\([Cc][Oo][Nn][tT][eE][Nn][tT][-][Ll][eE][nN][gG][tT][hH][:][ \t]*\\([0-9]+\\)\r\n\\)\\(\\(.*\r\n\\)*\\)";
    fields = [  4, Some "0" ] }

let transfer_coding_regexp =
  { regexp = Str.regexp "\\(\\(.*\r\n\\)*\\)\\([tT][rR][aA][nN][sS][fF][eE][rR]-[eE][Nn][cC][oO][dD][iI][nN][gG][:][ \t]*\\(.*\\)\r\n\\)\\(\\(.*\r\n\\)*\\)\r\n";
    fields = [  1, None; 4, Some "" ; 5, None  ] }

let e_tag_regexp =
  { regexp = Str.regexp "\\(\\(.*\r\n\\)*\\)\\([Ee][tT][aA][gG][:][ \t]*\\(W/\\)?[ \t]*\"\\(.*\\)\"[ \t]*\r\n\\)\\(\\(.*\r\n\\)*\\)";
    fields = [  5, Some "" ] }

let last_modified_regexp =
  { regexp = Str.regexp "\\(\\(.*\r\n\\)*\\)\\([Ll][aA][sS][tT]-[Mm][oO][dD][iI][fF][iI][Ee][Dd][:]\\(.*\\)\r\n\\)\\(\\(.*\r\n\\)*\\)";
    fields = [  4, Some "" ] }

let hop_by_hop_regexp =
{ regexp = Str.regexp "\\(\\(.*\r\n\\)*\\)\\(\\([tT][rR][aA][nN][sS][fF][eE][rR]-[eE][Nn][cC][oO][dD][iI][nN][gG]:.*\r\n\\)\\|\\([Cc][Oo][nN][nN][eE][cC][tT][iI][oO][nN][:].*\r\n\\)\\|\\([Tt][rR][aA][iI][lL][eE][rR][sS][:].*\r\n\\)\\|\\([Kk][Ee][eE][pP]-[Aa][lL][iI][vV][eE][:].*\r\n\\)\\)\\(\\(.*\r\n\\)*\\)";
    fields = [  1, None; 8, None ] }

let unfeatured_regexp =
{ regexp = Str.regexp "\\(\\(.*\r\n\\)*\\)\\(\\([Ii][fF]-[Mm][oO][dD][iI][fF][iI][eE][Dd]-[Ss][iI][nN][cC][eE][:].*\r\n\\)\\|\\([Ii][fF]-[Nn][oO][nN][eE]-[Mm][aA][tT][cC][Hh]:.*\r\n\\)\\)\\(\\(.*\r\n\\)*\\)";
    fields = [  1, None; 6, None ] }

let ok_to_write_regexp =
{ regexp = Str.regexp "\\(\\(.*\r\n\\)*\\)\\([Oo][Kk]-[tT][oO]-[Ww][rR][iI][tT][eE]\\)\\(\\(.*\r\n\\)*\\)";
    fields = [  3, Some "OK" ] }

let parse_cached_header string =
  match regexp_match ok_to_write_regexp string  with
  | Some ( headers_1 :: _ ) -> "OK"
  | None -> "Not_OK"
  | _ -> error_info "Strange things in cached header" string; raise Kill_connection;;

let parse_host_header string =
  match regexp_match host_header_regexp string  with
  | Some ( headers_1 :: headers_2 :: _ ) -> headers_1 ^ headers_2
  | None -> string
  | _ -> error_info "Ill formed headers (host)" string; raise Kill_connection;;

let parse_hop_by_hop string =
  match regexp_match hop_by_hop_regexp string  with
  | Some ( headers_1 :: headers_2 :: _ ) -> headers_1 ^ headers_2
  | None -> string
  | _ -> error_info "Ill formed headers (Hop by Hop)" string; raise Kill_connection;;

let parse_unfeatured string =
  match regexp_match unfeatured_regexp string  with
  | Some ( headers_1 :: headers_2 :: _ ) -> headers_1 ^ headers_2
  | None -> string
  | _ -> error_info "Ill formed headers (Unfeatured)" string; raise Kill_connection;;

let parse_size string =
  match regexp_match chunk_size_regexp string  with
  | Some ( size :: _ ) -> size
  | _ -> error_info "Bad Size in chunk" string; raise Kill_connection;;

let parse_content_length_header string =
  match regexp_match content_length_regexp string  with
  | Some ( length :: _ ) -> length
  | None -> "0"
  | _ -> error_info "Ill formed headers (content-length)" string; raise Kill_connection;;

let parse_transfer_coding_header string =
  match regexp_match transfer_coding_regexp string  with
  | Some ( headers_1 :: trans :: headers_2 :: __ ) ->  trans
  | None ->  ""
  | _ -> error_info "Ill formed headers (transfer-coding)" string; raise Kill_connection;;

let parse_e_tag_header string =
  match regexp_match e_tag_regexp string  with
  | Some ( etag :: _ ) -> etag
  | None -> ""
  | _ -> error_info "Ill formed headers (E-Tag)" string; raise Kill_connection;;

let parse_last_modified_header string =
  match regexp_match last_modified_regexp string  with
  | Some ( date :: _ ) -> date
  | None -> ""
  | _ -> error_info "Ill formed headers (Last Modified)" string; raise Kill_connection;;

let parse_request string =
  match regexp_match request_regexp string  with
  | Some ( request_line :: headers :: entity :: _ ) -> request_line, headers, entity
  | _ -> error_info "Ill formed request" string; raise Kill_connection;;

let parse_request_line string =
  match regexp_match request_line_regexp string  with
  | Some ( meth :: request_uri :: http_vers:: _ ) -> meth , request_uri , http_vers
  | _ -> error_info "Ill formed request-line" string; raise Kill_connection;;

let parse_status_line string =
  match regexp_match status_line_regexp string  with
  | Some (  http_vers:: status_code :: reason :: _ ) -> http_vers, status_code, reason
  | _ -> error_info "Ill formed status-line" string; raise Kill_connection;;

let parse_host host =
  match regexp_match host_regexp host with
    Some (host :: port :: _) -> host, int_of_string port
  | _ -> error_info "Ill formed host" host;raise Kill_connection;;

let parse_host_2 host =
  match regexp_match host_regexp_2 host with
    Some (host :: port :: _) -> host, port
  | _ -> error_info "Ill formed host" host;raise Kill_connection;;

let parse_url url =
  match regexp_match url_regexp url with
    Some (host :: path :: _) -> parse_host host, path
  | _ -> error_info "Ill formed url" url; raise Kill_connection;;

let parse_uri request_uri proxy =
    match proxy with
      None -> parse_url request_uri
    | Some host -> parse_host host, request_uri;;

let parse_uri_true request_uri =
  match regexp_match url_regexp request_uri with
    Some (host :: path :: _) -> parse_host_2 host, path
  | _ -> error_info "Ill formed url (2)" request_uri; raise Kill_connection;;
