(*

 Version 0.4

   Many features are now implemented (the main missing one is permanent connections). There is an somewhat advanced thread pool management; and the cache is more efficient (can't grow undefinitely, more robust). A lot of bugs on the communication functions have been fixed.
   There are of course many possible improvements, but at this point probably the more useful thing to do next is thoroughly test this proxy. For exemple, the whole program can easily be blocked if only one thread is affected (cause unknown...)


   Next: Testing and debugging... Persistent connections.

 *)

(****************************************************************)

open Sys;;
open Unix;;
open Thread;;
open Event;;

open Parsing;;

type options = { mutable proxy : string option; mutable verbose : bool; mutable debug : bool; mutable server : bool; mutable unsecure : bool; mutable max_threads : int; mutable min_spare : int;
 mutable max_spare : int; mutable port : int; mutable max_cache_size : int;  mutable max_file_size : int; mutable expiration_time : float; mutable server_dir : string; mutable cache_dir : string; mutable headers_cache_dir : string }

type vars = { mutable ready : int; mutable busy : int; mutable all : int }

type vars_2 = { mutable cache_size : int; mutable available_size : int}

type cache_chan = { chan : out_channel; mutable size : int}

exception Kill_connection;;
exception Terminate_thread;;
exception Non_cachable;;
exception No_cache;;
exception Not_in_cache;;
exception Bad_request_400;;
exception Not_found_404;;
exception Not_implemented_501;;

let threads_vars_lock = Mutex.create();;
let cache_vars_lock = Mutex.create();;
let sock_lock = Mutex.create();;

type chan =
    Standard of out_channel
  | Signal
  | Cache of cache_chan;;

let buffer_size = 10240;;
 
let global = { 
  proxy = None;
  verbose = false;
  debug = false;
  server = false;
  unsecure = false;
  max_threads = 40;
  min_spare = 10;
  max_spare = 20;
  port = 8080;
  max_cache_size = 5242880;
  max_file_size = 153600;
  expiration_time = 300.12;
  server_dir = "server";
  cache_dir = "cache";
  headers_cache_dir = "cache_headers";
};;

let cache_vars = {
  cache_size = 0;
  available_size = global.max_cache_size;
}

let threads_vars = {
  ready = 0;
  busy = 0;
  all = 0;
};;



(****************************************************************)

(********** General functions ****** **********)


let rec do_channel_list f g list = match list with
| [] -> ()
| Standard c::r -> f c; do_channel_list f g r
| Cache c::r-> g c; do_channel_list f g r
| Signal::r -> (); do_channel_list f g r;;

let do_dir f dirname = let d = opendir dirname in
try
  while true do f (readdir d)
  done
with End_of_file -> closedir d;;


(*** Different functions for input/output **********)


let send_request chan request =
  Pervasives.output_string chan request ;
  flush chan;;

let transfer n chan_in chan_out =
  let buffer = String.create n in
  let rec copy x =
    let k = input chan_in buffer 0 x in
    output chan_out buffer 0 k;
    if (k > 0)&(k<x) then copy (x-k) in
  copy n;
  flush chan_out;;
 
let transfer_unknown chan_in chan_out =
  let buffer = String.create buffer_size in
  let rec copy () =
    let n = input chan_in buffer 0 buffer_size in
    output chan_out buffer 0 n;
    if n > 0 then copy () in
  copy();
  flush chan_out;;

let transfer_unto_string chan_in buffer =
  let m = String.length buffer in
  let rec copy x =
    let n = input chan_in buffer x (m-x) in
    if n > 0 then copy (x+n) in
  copy 0;;

let multi_transfer n chan_in channels_out =
  let buffer = String.create n in 
  let rec copy x =
    let k = input chan_in buffer 0 x in   
    flush Pervasives.stdout;
    let f c = output c buffer 0 k in
    let g c = if (k <= c.size) then 
      output c.chan buffer 0 k else c.size <- -1 in
    do_channel_list f g channels_out;
    if (k > 0) & (k<x) then copy (x-k) in
  copy n;
  let f c = flush c in let g c = flush c.chan in
  do_channel_list f g channels_out;;

let multi_transfer_unknown chan_in channels_out =
  let buffer = String.create buffer_size in
  let rec copy z =
    let n = input chan_in buffer 0 buffer_size in 
    let f c = output c buffer 0 n in
    let g c = if (n <= c.size) then 
      output c.chan buffer 0 n else (c.size <- -1) in
    do_channel_list f g channels_out;
    if n > 0 then copy (z+n) else begin
  let f c = flush c in let g c = flush c.chan in
  do_channel_list f g channels_out;
      z
    end;
  in
  copy 0;;


let rec get_line in_chan = 
  let line = input_line in_chan in
  if line = "\r" then get_line in_chan else line;;

let get_status_line in_chan = get_line in_chan;;

let get_request_line in_chan = get_line in_chan;;

let rec get_headers in_chan line = 
  let new_line = input_line in_chan in
  if new_line = "\r" then line else
  get_headers in_chan ( line ^ new_line ^ "\n");;


let rec get_chunks chan_in channels_out =
  let rec get_chunks_aux total_size =
    let new_line = get_line chan_in in 
      let f c = send_request c (new_line ^ "\n") in
      let g c s = () in
      do_channel_list f g channels_out;
    let size = int_of_string ( "0x" ^ (parse_size new_line ) ) in 
    if (size != 0) then begin
      multi_transfer size chan_in channels_out;
      let f c = transfer 2 chan_in c in
      let g c s = () in
      do_channel_list f g channels_out;
      get_chunks_aux (total_size+size);
    end
    else total_size;
  in get_chunks_aux 0;;


(****** Some miscellaenous functions *******)

let add_host_header headers request_uri =
let (host, port), trash= parse_uri_true request_uri in
let line= if port = "" then host else (host ^ ":" ^ port) in
"Host: " ^ line ^ "\r\n" ^ ( parse_host_header headers);;

let add_etag_validator etag msg = (msg ^ "If-None-Match: " ^ "\"" ^ etag ^ "\"\r\n");;

let add_mod_validator date msg = (msg ^ "If-Modified-Since: " ^ date ^ "\r\n");;

let connect_to_server hostname port  =
 let hostaddr =
    try inet_addr_of_string hostname
    with Failure _ -> try (gethostbyname hostname).h_addr_list.(0) 
	with x -> error_info "DNS Error..." ""; raise Kill_connection
in
open_connection (ADDR_INET (hostaddr, port));;


(*********** Different functions dealing with cache management on disk ***************)

let encode_path path = let n = String.length path in if n > 100 then raise No_cache
    else
for i = 0 to (n-1) do
if String.get path i = '/' then String.set path i '_' done;
path;;

let url_to_cache hostname port path = let hdir = Filename.concat global.headers_cache_dir (hostname ^ "_" ^ port ) in
let dir = Filename.concat global.cache_dir (hostname ^ "_" ^ port ) in
hdir, dir, (encode_path path);;

let make_cache_file dir file = 
 begin
  try
    let trash = stat dir in ()
  with Unix_error (ENOENT,_,_) -> mkdir dir 0o777
  end;
let cached_file = Filename.concat dir file in 
  out_channel_of_descr (openfile cached_file (O_WRONLY::O_CREAT::O_TRUNC::[]) 0o666);;



let check_cache dir file =
let cached_file = Filename.concat dir file in
try
openfile cached_file (O_RDONLY::[])  0o666;
with _ -> raise Not_in_cache;;


let create_cache_file hdir cdir file = 
 begin
  try
    let trash = stat cdir in ()
  with Unix_error (ENOENT,_,_) -> mkdir cdir 0o777
 end;
  let cached_file = Filename.concat cdir file in
  

  try begin
    let infos = stat cached_file in 
    let cached_header = Filename.concat hdir file in
    try begin
      let infos_2 = stat cached_header in 
      let buffer = String.create infos_2.st_size in
      let chan_in = in_channel_of_descr (openfile cached_header (O_RDONLY::[]) 0o666) in
      transfer_unto_string chan_in buffer;
      if not ((parse_cached_header buffer) = "OK") then cached_file, Signal
      else let desc = openfile cached_file (O_WRONLY::O_CREAT::O_TRUNC::[]) 0o666 in
      Mutex.lock cache_vars_lock;
      let channel = Cache { chan= out_channel_of_descr desc; size = min cache_vars.available_size global.max_file_size } in
      Mutex.unlock cache_vars_lock;
      cached_file, channel;
    end;
    with Unix_error (ENOENT,_,_) -> cached_file, Signal ;
  end;
  with Unix_error (ENOENT,_,_) -> 


    let desc = openfile cached_file (O_WRONLY::O_CREAT::O_TRUNC::[]) 0o666 in
  
    Mutex.lock cache_vars_lock;

    let channel = Cache { chan= out_channel_of_descr desc; size = min cache_vars.available_size global.max_file_size } in

    Mutex.unlock cache_vars_lock;
    cached_file, channel;;

let count_cache_size () = 
  cache_vars.available_size <- global.max_cache_size;
  let rec f source file = if file <> Filename.current_dir_name && file <> Filename.parent_dir_name then 
    count_rec ( Filename.concat source file) and 
      count_rec source =
    let infos = stat source in
    match infos.st_kind with 
      S_REG -> cache_vars.available_size <- cache_vars.available_size - infos.st_size;
    | S_DIR -> do_dir (f source )source;
    | _ -> error "Major problems encountered in the cache directories.\n" "Please check them and purge them manually.\n";
  in
  count_rec global.cache_dir;;

let clean_cache () = Mutex.lock cache_vars_lock;
    let rec f source file = if file <> Filename.current_dir_name && file <> Filename.parent_dir_name then
      clean_rec ( Filename.concat source file) and	
	clean_rec source =
      let infos = stat source in
      match infos.st_kind with
	S_REG -> if (infos.st_atime > global.expiration_time) && (infos.st_mtime > global.expiration_time) then
	  unlink source; unlink (global.headers_cache_dir ^ (String.sub source (String.length global.cache_dir) (String.length source - (String.length global.cache_dir))));
      | S_DIR -> do_dir (f source) source;
      | _ -> error "Major problems encountered in the cache directories.\n" "Please check them and purge them manually.\n";
    in   
    clean_rec global.cache_dir;
    count_cache_size ();
      Mutex.unlock cache_vars_lock;;

(***** These are functions when we act as a server (including when we serve from cache...) *******)

let serve_from_cache f_desc chan_out =

let stats = fstat f_desc in
let n = stats.st_size in
let chan_in = in_channel_of_descr f_desc in
output_string chan_out "HTTP/1.1 200 OK\r\n";
Printf.fprintf chan_out "Content-length: %i" n;
output_string chan_out "\r\nConnection: close\r\n";
output_string chan_out "\r\n";
transfer n chan_in chan_out;
close_in chan_in;;

let serve_file f_desc http_vers chan_out client_chan_in = 

  let chan_in = in_channel_of_descr f_desc in

  match http_vers with 
    "HTTP/0.9" -> transfer_unknown chan_in chan_out;
  | "HTTP/1.0"|"HTTP/1.1" -> let stats = fstat f_desc in 
    let n = stats.st_size in
    Printf.fprintf chan_out "%s 200 OK\r\nContent-length: %i" http_vers n;
    output_string chan_out "\r\nConnection: close\r\n\r\n";    let n = stats.st_size in
    transfer n chan_in chan_out;
  | _ -> raise Bad_request_400;;
  


(**************** This function transfers the document from the server to the cache and to the client ***)

let process_chunks reply chan_in channel_out hdir cdir file =

  output_string channel_out reply;
  print_string ("Chunked encoding.\n");
  flush Pervasives.stdout; 
      
  let cache_file, channel = create_cache_file hdir cdir file in   
  let a = get_chunks chan_in ((Standard channel_out)::channel::[])  in
  let entity_headers =  get_headers chan_in "" in
  output_string channel_out (entity_headers ^ "\r\n");
  begin
    match channel with
    | Standard _ | Signal -> ()
    | Cache c -> 	close_out c.chan;

      if (c.size >= 0) then begin

	let h_channel = make_cache_file hdir file in	
	output_string h_channel reply;
	output_string h_channel (entity_headers ^ "\r\n");
	output_string h_channel ("OK-to-Write\r\n\r\n");
	close_out h_channel;
	
	Mutex.lock cache_vars_lock;
	cache_vars.available_size <- (cache_vars.available_size - a);
	Mutex.unlock cache_vars_lock
      end
	  
  else begin 
    unlink cache_file;
    clean_cache (); 
  end;
  end;;


let rec transfer_from_server reply chan_in channel_out hdir cdir file =

  match file with
| None ->  

    output_string channel_out reply;
    let tcoding = parse_transfer_coding_header reply in
    if (tcoding = "chunked") then begin  print_string ("Chunked encoding.\n");
    flush Pervasives.stdout; 
    ignore (get_chunks chan_in ((Standard channel_out)::[]));
    end
    else
      
      let k = int_of_string (parse_content_length_header reply) in
      if k> 0 then begin
	print_string ("Content-length encoding.\n");
	flush Pervasives.stdout;
	transfer k chan_in channel_out;	
      end
      else 
	print_string ("Unknown Encoding.\n");
      flush Pervasives.stdout;
      transfer_unknown chan_in channel_out;

| Some file ->

let tcoding = parse_transfer_coding_header reply in

    if (tcoding = "chunked") then process_chunks reply chan_in channel_out hdir cdir file

    else begin  
      let k = int_of_string (parse_content_length_header reply) in
      if k> 0 then begin
	print_string ("Content-length encoding.\n");
	flush Pervasives.stdout;
	Mutex.lock cache_vars_lock;
	let b = (k > cache_vars.available_size) in
	Mutex.unlock cache_vars_lock;	
	if b then clean_cache ();    	
	Mutex.lock cache_vars_lock;
	let c = (k <= global.max_file_size) && (k <= cache_vars.available_size) in
	Mutex.unlock cache_vars_lock;
		
	if c then begin
	  let cache_file, channel = create_cache_file hdir cdir file in 
	  match channel with
	  | Standard _ | Signal -> transfer_from_server reply chan_in channel_out hdir cdir None;
	  | Cache c -> 

	      output_string channel_out reply;
	      multi_transfer k chan_in ((Standard channel_out)::channel::[]);    
	      close_out c.chan;    
	      Mutex.lock cache_vars_lock;
	      cache_vars.available_size <-  cache_vars.available_size - k;
	      Mutex.unlock cache_vars_lock;
	      let h_chan = make_cache_file hdir file in 
	      output_string h_chan reply;
	      output_string h_chan ("OK-to-Write\r\n\r\n");
	      close_out h_chan;
	end

	else transfer_from_server reply chan_in channel_out hdir cdir None;
	
      end

      else begin
	print_string ("Unknown Encoding.\n");
	flush Pervasives.stdout;
	output_string channel_out reply;
      let cache_file, channel = create_cache_file hdir cdir file in 
      let a = multi_transfer_unknown chan_in ((Standard channel_out)::channel::[]) in
    match channel with
    | Standard _ | Signal  -> ()
    | Cache c -> close_out c.chan;
	if (c.size >= 0) then begin
	  let h_chan = make_cache_file hdir file in 
	  output_string h_chan reply;
	  output_string h_chan ("OK-to-Write\r\n\r\n");
	  close_out h_chan;
	  Mutex.lock cache_vars_lock;
	  cache_vars.available_size <-  cache_vars.available_size - a;
	  Mutex.unlock cache_vars_lock;  
	end	
	else begin
	  unlink cache_file;
	  clean_cache ();
	end;
      end;
    end;;

let a = 3;;

(********* This function checks for any validators associated with the cached file ***********)

let prepare_validation f_desc msg = 
let cache_out_chan = in_channel_of_descr f_desc in
let st_line = get_status_line cache_out_chan in
let trash, status_code, reason = parse_status_line st_line in
let reply_headers = get_headers cache_out_chan "" in
(*get validators *)
let etag = parse_e_tag_header reply_headers in
let msg, valid = if etag = "" then  msg, false else (add_etag_validator etag msg), true in 
let lmod = parse_last_modified_header reply_headers in
let msg, valid = if lmod = "" then msg, valid else (add_mod_validator lmod msg), true in
if valid=false then print_string ("No validators found. We won't use cache for this request.\n"); 
flush Pervasives.stdout; 
msg;;

(************ We remove hop-by-hop headers and unfeatured headers (such as those controling cache for the client...) **********)

let treat_headers headers = 
  let rec remove_hop_by_hop str =
    let str2 = parse_hop_by_hop str in 
    if str2 = str then str else remove_hop_by_hop str2  
  in 
  let headers2 = remove_hop_by_hop headers in

  let rec remove_unfeatured str =
    let str2 = parse_unfeatured str in 
    if str2 = str then str else remove_unfeatured str2   
  in let headers3 = remove_unfeatured headers2
  in (headers3 ^ "Connection: close\r\n");;


(***** Preparing forwarded request... ******************)

let make_fwd_request meth path request_uri http_vers in_chan =

match http_vers with
 |"HTTP/0.9" -> (meth ^ " " ^ path ^ " "  ^ "\r\n")
 |"HTTP/1.0" -> let headers = treat_headers (get_headers in_chan "" ) in 
   if global.debug then begin
     print_string headers;
     flush Pervasives.stdout; 
   end;
   (meth ^ " " ^ path ^ " " ^ " " ^ http_vers ^ " \r\n" ^ headers) 
 |"HTTP/1.1" ->	let headers = treat_headers (get_headers in_chan "" ) in 
   if global.debug then begin
     print_string headers;
     flush Pervasives.stdout; 
   end;
   (meth ^ " " ^ path ^ " " ^ " " ^ http_vers ^ " \r\n" ^ (add_host_header headers request_uri))
 | _ -> error_info "Unrecognized HTTP version:" http_vers; raise Kill_connection;;


let send_full_request hostname port request client_out_chan =

(* Connection to server. We get headers and status-line. *)

let server_out_chan, server_in_chan = connect_to_server hostname port in
send_request server_in_chan ( request ^ "\r\n" );

if global.verbose then begin
print_string ("We sent the following request:\n****************************************\n" ^ request ^ "****************************************\n");
flush Pervasives.stdout; 
end;


let k = int_of_string (parse_content_length_header request) in
if k > 0 then transfer k client_out_chan server_in_chan;
let st_line = get_status_line server_out_chan in
let trash, status_code, reason = parse_status_line st_line in
let reply_headers = get_headers server_out_chan "" in

if global.verbose then begin
print_string ("We received the following headers:\n****************************************\n" ^ st_line ^ "\n" ^ reply_headers ^ "****************************************\n");
flush Pervasives.stdout 
end;
st_line, server_out_chan, status_code, reason, reply_headers ;;


let treat_without_cache request client_in_chan client_out_chan hostname port = 
  let st_line, server_out_chan, status_code, reason, reply_headers = send_full_request hostname port request client_out_chan in
      
  transfer_from_server (st_line ^ "\n" ^ reply_headers ^ "\r\n" ) server_out_chan  client_in_chan "" "" None;
  shutdown_connection server_out_chan;
  close_out client_in_chan;
  print_string ("\nRequest completed!\n\n");
  flush Pervasives.stdout;;

let treat_with_cache request client_in_chan client_out_chan hostname port hdir cdir cfile =

  let st_line, server_out_chan, status_code, reason, reply_headers = send_full_request hostname port request client_out_chan in

(*    let server_out_chan, server_in_chan, status_code, reason, reply_headers = send_full_request hostname port request in *)
    
    begin

      try
	match status_code with

	| "304" ->   print_string("We serve from cache the requested file.\n");
	    flush Pervasives.stdout;
	    serve_from_cache (check_cache cdir cfile) client_in_chan
	| "200" -> 
	  print_string ("We try to cache the answer.\n");

	  flush Pervasives.stdout;     
	  transfer_from_server (st_line ^ "\n" ^ reply_headers ^ "\r\n") server_out_chan client_in_chan hdir cdir (Some cfile);

	| _ -> raise Non_cachable;
		
      with Non_cachable ->
  
	print_string ("Not cachable\n");
	flush Pervasives.stdout;
	transfer_from_server (st_line ^ "\n" ^ reply_headers ^ "\r\n" ) server_out_chan  client_in_chan "" "" None;
	
    end;  

(* The server has now replied, and request is normally fulfilled *)

shutdown_connection server_out_chan;
close_out client_in_chan;
print_string ("\nRequest completed!\n\n");
flush Pervasives.stdout;;


(***********************************************************)

(********************* Beginning of main thread *****************************************)

let http_proxy client_out_chan client_in_chan =

try

let line= get_request_line client_out_chan in

if global.debug then begin
print_string ("We received the following raw request:\n" ^ line ^ "\n");
flush Pervasives.stdout 
end;

let meth , request_uri , http_vers = parse_request_line line in 
if not(meth = "GET") then begin
  print_string "This proxy only supports GET requests, sorry.";
  flush Pervasives.stdout;
  if not global.unsecure then raise Kill_connection
end;
let (hostname, port), path = parse_uri request_uri global.proxy in
let request = make_fwd_request meth path request_uri http_vers client_out_chan in


print_string ("Request understood. Attempting to reply...\n");
flush Pervasives.stdout;

(* Reply *)

match http_vers with

(* we are here dealing with http 0.9 *)
(* we won't use cache with http 0.9 *)

| "HTTP/0.9" -> let server_out_chan, server_in_chan = connect_to_server hostname port in

  send_request server_in_chan request;
transfer_unknown server_out_chan client_in_chan;

(* now http 1.1 *)

| _ -> let (host, p), path = parse_uri_true request_uri in

  try
    let hdir, cdir, cfile = url_to_cache host p path in
    let request =
      begin
	try begin
	  let f_desc = check_cache hdir cfile in

(* we MUST revalidate cache first *)

	  print_string ("Cached file found. Attempting to validate with origin server...\n");
	  flush Pervasives.stdout;
	  try    
	    prepare_validation f_desc request;     
	  with _ -> error_info "There was a problem with the cached file. We won't use cache for this request." ""; raise Not_in_cache
	end;
	with Not_in_cache -> request
      end;
    
    in 

if global.debug then begin
print_string ("OK. We now send request to server.\n");
flush Pervasives.stdout 
end;

treat_with_cache request client_in_chan client_out_chan hostname port hdir cdir cfile;
  
  with No_cache -> 
    print_string ("It seems there will be problems if we want to use cache. Cache disabled for request.\nName of file:");
    print_string (path ^ "\n");
    flush Pervasives.stdout;

treat_without_cache request client_in_chan client_out_chan hostname port;

with End_of_file ->
  error_info "Ill formed request" "End_of_file encountered"; raise Kill_connection
| Kill_connection -> raise  Kill_connection;;



(*****************************************************************)
(**************** Server implementation **************************)



let rec get_path string =
  if not((String.get string 0) = '/') then raise Bad_request_400 else let n= String.length string in
  if n = 1 then "" else

(* for i = 1 to (n-1) do *)

  let rec advance str i =

    if (i = String.length str) then String.sub str 1 (n-1) else
    
    if ((String.get str i) = '/') then
      Filename.concat (String.sub str 1 (i-1)) (get_path (String.sub str i (n-i)) )
    else advance str (i+1)
  in advance string 1;;

let http_server client_out_chan client_in_chan =

try
  

  begin
  try 
    let line= get_request_line client_out_chan in

    let meth , path , http_vers = parse_request_line line in 

    if global.verbose then begin
      print_string ( line  ^ "\n" );
      flush Pervasives.stdout
    end;
    
    let headers = 
      match http_vers with 
	"HTTP/0.9" -> ""
      | "HTTP/1.0"|"HTTP/1.1" -> get_headers client_out_chan ""
      | _ -> raise Bad_request_400  
    in
    let k = int_of_string (parse_content_length_header headers) in
    if k > 0 then begin
      let buffer = String.create (k+10) in

    let rec copy i m = 
      let n = input client_out_chan buffer i m in
      if n> 0 then copy (i+n) (m-n)
    in copy 0 k;
    end;
    
    if global.verbose then begin
      print_string headers;
      flush Pervasives.stdout
    end;

    if not(meth = "GET") then begin
      print_string "This server only supports GET requests, sorry.";
      flush Pervasives.stdout; raise Not_implemented_501
    end;


    let file = Filename.concat global.server_dir (get_path path) in 

    let file2 =
      let st = stat file in
      if st.st_kind = S_DIR then (Filename.concat file "index.html") else file
    in serve_file (openfile file2 (O_RDONLY::[]) 0o666) http_vers client_in_chan client_out_chan;
	

  with Unix_error (ENOENT,_,_) | Not_found_404 -> output_string client_in_chan "HTTP/1.1 404 Not Found\r\nConnection: close\r\nContent-length: 0\r\n\r\n";
		     
| Bad_request_400 -> output_string client_in_chan "HTTP/1.1 400 Bad Request\r\nConnection: close\r\n\r\n";
| Not_implemented_501 -> output_string client_in_chan "HTTP/1.1 501 Not Implemented\r\nConnection: close\r\n\r\n";
| Kill_connection -> output_string client_in_chan "HTTP/1.1 500 Internal Server Error\r\nConnection: close\r\n\r\n";
| End_of_file -> error_info "Ill formed request" "End_of_file encountered"; raise End_of_file
| x -> raise x
      
  end;
  
  close_out client_in_chan;
  print_string ("\nRequest completed.\n\n");
  flush Pervasives.stdout;

with x -> raise x;;

(****************************************** Initialization ****************************************)

let wrap_channel f s =
  begin
  try
    let reply = in_channel_of_descr s in
    let request = out_channel_of_descr s in
    f reply request;
    close_out request;
  with Kill_connection ->  print_string "Connection killed.\n"; close s;
  | x -> raise x;
  end;
flush Pervasives.stdout;;

let rec return_from_request sock = 

Mutex.lock threads_vars_lock;
try
  begin
    threads_vars.busy <- threads_vars.busy - 1;
    
    if threads_vars.ready >= global.max_spare then begin
      
      threads_vars.all <- threads_vars.all - 1;     
      raise Terminate_thread;
	
    end;

    threads_vars.ready <- threads_vars.ready + 1;

    if global.verbose then begin
      print_string "Status of threads pool:\n";   
      print_string "Ready: ";
      print_int threads_vars.ready;
      print_string "\n";
      print_string "Busy: ";
      print_int threads_vars.busy;
      print_string "\n"; 
      print_string "All: ";
      print_int threads_vars.all;
      print_string "\n";
      flush Pervasives.stdout;
  end;
    
    if (threads_vars.ready < global.min_spare) && (threads_vars.all < global.max_threads) then begin
      
      threads_vars.all <- threads_vars.all + 1;    
      threads_vars.busy <- threads_vars.busy + 1;
      let th = create proxy_thread sock  in ();
      
    end;
    
    Mutex.unlock threads_vars_lock;
  end;
with x -> Mutex.unlock threads_vars_lock; raise x;

and

 go_into_request sock = 

Mutex.lock threads_vars_lock;

try
  begin
    threads_vars.busy <- threads_vars.busy + 1;
    if (threads_vars.ready <= global.min_spare && threads_vars.all < global.max_threads) then begin
      threads_vars.all <- threads_vars.all + 1; 
      threads_vars.busy <- threads_vars.busy + 1; 
      let th = create proxy_thread sock in ();	
    end;
    threads_vars.ready <- threads_vars.ready - 1; 
    
    Mutex.unlock threads_vars_lock;
  end;
with x -> Mutex.unlock threads_vars_lock; raise x;


and 

    proxy_thread sock =   

  try
    begin
      while true do

      return_from_request sock;
	
      let s = 
      
	Mutex.lock sock_lock;
	
	try
	  let (sk, hostaddr) = accept sock in Mutex.unlock sock_lock; sk
	with x-> Mutex.unlock sock_lock; raise x;

      in
	
	
      print_string "Incoming connection.\n";
      flush Pervasives.stdout;
    
      go_into_request sock ; 
      
      if global.server then
	wrap_channel http_server s
      else
	wrap_channel http_proxy s
    done;
    end;
    
  with Terminate_thread -> ();
  | x -> error_info "Unknown Exception has killed thread..." ""; 
      
(* Here we assume that the error occured when we were treating a connection (Reasonable assumption). *)
(* So we update the values for the threads according to that *)
      
      Mutex.lock threads_vars_lock;
      threads_vars.all <- threads_vars.all - 1; 
      threads_vars.busy <- threads_vars.busy - 1;   
      Mutex.unlock threads_vars_lock;
      Mutex.unlock sock_lock;
      raise x;;


let establish_coserver sockaddr =
  let socket = socket PF_INET SOCK_STREAM 0 in
  setsockopt socket SO_REUSEADDR true;
  bind socket sockaddr;
  listen socket 3;
  threads_vars.all <- 1;
  threads_vars.busy <- 1;
  
  let th = create proxy_thread socket in ();
  while true do sleep 1000 done;;

let proxy () =
  
  let rec parse_args i array =

    if (i < Array.length array) then
      
      match array.(i) with
      |  "-p" | "-port" -> begin
	  try global.port <- (int_of_string array.(i+1) )
	  with Invalid_argument _ -> error "Incorrect port" array.(i+1)
      end;
	  parse_args (i+2) array
      | "-v" | "-verbose" -> global.verbose <- true; parse_args (i+1) array
      | "-d" | "-debug" -> global.debug <- true; parse_args (i+1) array
      | "-s" | "-server" -> global.server <- true; parse_args (i+1) array 
      | "-u" | "-unsecure" -> global.unsecure <- true; parse_args (i+1) array
      | "-c" | "-cache" -> begin
	  try global.max_cache_size <- (int_of_string array.(i+1) )
	  with Invalid_argument _ -> error "Incorrect cache-size" array.(i+1)
      end;
      | "-sd" | "-serverdir" -> 
	  global.server_dir <- array.(i+1);	  
	  parse_args (i+2) array
      | "-t" | "-threads" -> begin
	  try global.max_threads <- (int_of_string array.(i+1) )
	  with Invalid_argument _ -> error "Incorrect number of threads" array.(i+1)
      end;      
      | "-mns" | "-minspare" -> begin
	  try global.min_spare <- (int_of_string array.(i+1) )
	  with Invalid_argument _ -> error "Incorrect number for min-spare" array.(i+1)
      end; 
      | "-mxs" | "-maxspare" -> begin
	  try global.max_spare <- (int_of_string array.(i+1) )
	  with Invalid_argument _ -> error "Incorrect number for max-spare" array.(i+1)
      end;
	  parse_args (i+2) array
      | "-r" | "-redirect" -> global.proxy <- Some array.(i+1);  parse_args (i+2) array
      | _ -> error "Unknown option:" array.(i)
	    
  in parse_args 1 Sys.argv;
  count_cache_size (); 
  
  if global.verbose then begin
    print_string "Available space for the cache: ";
    print_int cache_vars.available_size;
    print_string "\n";
    flush Pervasives.stdout; 
  end;
  
count_cache_size ();
  
  let http_port = global.port in 
  let my_host_addr = (gethostbyname (gethostname())).h_addr_list.(0) in
  let sockaddr = ADDR_INET(my_host_addr, http_port) in
  establish_coserver sockaddr;;


handle_unix_error proxy ();;

